#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

from dap_core import common
from dap_core.util import str_util


class DapActionBase:

  def run(self):
    common.print_info("Running " + self.name() + " action")

  def name(self):
    class_name = self.__class__.__name__
    return str_util.camel_to_snake(class_name.replace("Action", ""))



#!/usr/bin/env python

#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

import sys, pathlib

sys.path.append(str(pathlib.Path(__file__).parent.parent.absolute()))

from dap_core import common, actions
from dap_core.meta_model import MetaModel


def run(log_level="info", meta_model_name="power_daps/python3", actions_to_run=["default"]):
  common.set_log_level(log_level)
  meta_model = MetaModel(meta_model_name)
  common.set_meta_model(meta_model_name)

  valid_actions = meta_model.actions()
  valid_actions += actions.local_actions()
  valid_action_names = [common.action_name(va) for va in valid_actions]

  common.print_verbose('Actions to run ' + str(actions_to_run))
  common.print_verbose('Valid actions ' + str(valid_action_names))

  for action_to_run in actions_to_run:
    if action_to_run not in valid_action_names:
      common.print_error("Action '" + action_to_run + "' not found.")
      continue
    for valid_action in valid_actions:
      valid_action_name = common.action_name(valid_action)
      if valid_action_name == action_to_run:
        common.stop_if_failed(*valid_action.run())


def main():
  import argparse
  import os
  default_meta_model = common.configured_meta_model()

  parser = argparse.ArgumentParser(description="dap")
  parser.add_argument("-v", "--verbose", dest="log_level",
                      default="info", action="store_const",
                      const="verbose", help="Verbose output")
  parser.add_argument("-q", "--quiet", dest="log_level",
                      default="info", action="store_const",
                      const="error", help="Quiet output. Only errors will be written out.")
  parser.add_argument("-m", "--meta-model", dest="meta_model",
                      default=default_meta_model,
                      help="Use the specified meta-model. Defaults to 'power_daps/python3'. " +
                           "Options are: power_daps/python3, power_daps/rust, power_daps/java and " +
                           "power_daps/es6. You can also just use 'python3', 'rust', 'java' or 'es6'. " +
                           "The 'power_daps/' will automatically get added.")
  parser.add_argument("action",
                      help="List of actions to run. Defaults to 'default' for the given meta-model",
                      default=["default"], nargs="*")

  args = parser.parse_args()

  run(args.log_level, args.meta_model, args.action)
  sys.exit(common.exit_code())


if __name__ == '__main__':
  main()



#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

from dap_core import common, dependencies
from dap_core.base_actions.dap_action_base import DapActionBase


class DepsActionBase(DapActionBase):
  default_dependencies_file_location = common.dependencies_file_location()

  def __init__(self, dependencies_file_location=""):
    super().__init__()
    self.set_dependencies_file_location(dependencies_file_location)

  def run(self):
    super().run()
    self.deps = self.load_dependencies(self.dependencies_file_location)

    for dep in self.deps.dependencies_for("default"):
      dep.install()

    return 0, ""

  def load_dependencies(self, dependencies_file_location):
    dependencies_file_contents = ""
    try:
      with open(dependencies_file_location) as f:
        dependencies_file_contents = f.read()
      f.closed
    except (OSError, IOError) as e:
      common.print_warning("No dependencies file found at " + dependencies_file_location)

    return dependencies.Dependencies(dependencies_file_contents)

  def set_dependencies_file_location(self, dependencies_file_location=""):
    if dependencies_file_location:
      self.dependencies_file_location = dependencies_file_location
    else:
      self.dependencies_file_location = DepsActionBase.default_dependencies_file_location





#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

from dap_core import common
from dap_core.meta_model import MetaModel
from dap_core.no_action_error import NoActionError
from dap_core.base_actions.dap_action_base import DapActionBase
import yaml


class DefaultActionBase(DapActionBase):
  default_actions_file_location = common.actions_file_location()

  def __init__(self, actions_file_location="./actions.yml"):
    super().__init__()
    self.set_actions_file_location(actions_file_location)

  def run(self):
    super().run()

    with open(self.actions_file_location) as f:
      actions_file_contents = f.read()
      for stage in yaml.load(actions_file_contents, Loader=yaml.SafeLoader).items():
        if stage[0] == 'default':
          for an_action in stage[1]:
            common.stop_if_failed(*self.action_for(an_action).run())
    f.closed
    return 0, ""

  def action_for(self, action_name):
    meta_model = MetaModel(common.meta_model())
    the_actions = list(filter(lambda a: common.action_name(a) == action_name, meta_model.actions()))
    if the_actions:
      return the_actions[0]
    else:
      return NoActionError(action_name)

  def set_actions_file_location(self, actions_file_location):
    if actions_file_location:
      self.actions_file_location = actions_file_location
    else:
      self.actions_file_location = DefaultActionBase.default_actions_file_location





#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

import os, shutil
from dap_core import common


def setup_git(d):
  git_init(d)
  add_to_git(d)


def git_init(d):
  os.chdir(d)
  git_path = shutil.which('git')
  git_init_command = [git_path, 'init']
  common.run_command(git_init_command)


def git_init_or_add(d):
  if not is_in_git_repo(d):
    git_init(d)
  add_to_git(d)


def is_in_git_repo(d):
  os.chdir(d)
  exit_code, output = common.run_command_in_shell_without_output([common.which('git'), 'rev-parse', '--git-dir'])
  if exit_code == 0:
    return True
  else:
    return False


def add_to_git(d):
  os.chdir(d)
  git_path = shutil.which('git')
  git_add_command = [git_path, 'add', '.']
  common.run_command(git_add_command)

  git_commit_command = [git_path, 'commit', '-m', 'Initialized with power_daps template ' + common.meta_model()]
  common.run_command(git_commit_command)

  return 0, ""



#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

import yaml
from dap_core import common
from dap_core.meta_model import MetaModel
from dap_core.no_action_error import NoActionError
from dap_core.base_actions.dap_action_base import DapActionBase


class LocalAction(DapActionBase):
  def __init__(self, name, actions={}):
    super().__init__()
    self.actions = actions
    self.n = name

  def name(self):
    return self.n

  def run(self):
    super().run()
    exit_code = 0
    output = ""
    for a in self.actions:
      ec, o = action_for(a).run()
      exit_code += ec
      output += "\n" + output

    return exit_code, output


def local_actions():
  actions_file_location = "./actions.yml"
  actions = []
  try:
    with open(actions_file_location) as f:
      actions_file_contents = f.read()
      for stage in yaml.load(actions_file_contents, Loader=yaml.SafeLoader).items():
        action_name = stage[0]
        list_of_sub_actions = stage[1]
        if action_name != "default":
          actions.append(LocalAction(action_name, list_of_sub_actions))
    f.closed
  except(OSError, IOError) as e:
    common.print_verbose("No actions file found at " + actions_file_location)

  return actions


def action_for(action_name):
  meta_model = MetaModel(common.meta_model())
  the_actions = list(filter(lambda a: common.action_name(a) == action_name, meta_model.actions()))
  if the_actions:
    return the_actions[0]
  else:
    return NoActionError(action_name)



#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

import os, sys, re
import urllib.request
from distutils.version import LooseVersion
from xml.etree import ElementTree
from shutil import which
from dap_core import common


class CommandLineInstaller:
  def __init__(self, command_base):
    self.command_base = command_base
    return

  def install(self, dep_name, dep_version, details):
    exit_code, output = common.run_command(self.command_base + [dep_name])
    common.stop_if_failed(exit_code, output)


class PipInstaller:
  NOT_INSTALLED = 0
  SAME_VERSION_INSTALLED = 1
  OLDER_VERSION_INSTALLED = 2
  NEWER_VERSION_INSTALLED = 3

  def __init__(self):
    return

  def install(self, dep_name, dep_version="latest", details={}):
    package_name = dep_name
    if not dep_version == "latest":
      package_name = dep_name + "==" + str(dep_version)

    status = self.is_already_installed(dep_name, dep_version)
    if status == PipInstaller.NOT_INSTALLED:
      common.print_verbose(dep_name + " not installed. Installing.")
      command_to_run = [which('pip3'), '-q', 'install', package_name]
      exit_code, output = common.run_command(command_to_run)
      common.stop_if_failed(exit_code, output)

    elif status == PipInstaller.OLDER_VERSION_INSTALLED:
      common.print_verbose(dep_name + " is already installed. Upgrading to " + dep_version + " version.")

      command_to_run = [which('pip3'), '-q', 'install', '--upgrade', package_name]
      exit_code, output = common.run_command(command_to_run)
      common.stop_if_failed(exit_code, output)

    elif status == PipInstaller.NEWER_VERSION_INSTALLED:
      common.print_verbose(
        "Newer version of " + dep_name + " installed. Uninstalling and installing " + dep_version + " version.")

      command_to_run = [which('pip3'), '-q', 'uninstall', package_name]
      exit_code, output = common.run_command(command_to_run)
      common.stop_if_failed(exit_code, output)

      command_to_run = [which('pip3'), '-q', 'install', package_name]
      exit_code, output = common.run_command(command_to_run)
      common.stop_if_failed(exit_code, output)
    else:
      common.print_verbose(dep_name + ", " + dep_version + " is already installed. Doing nothing.")

  def is_already_installed(self, dep_name, dep_version):
    exit_code, output = common.run_command_in_shell("pip3 list --format=columns | grep -i " + dep_name)
    if not output:
      return PipInstaller.NOT_INSTALLED
    else:
      installed_version = re.sub(' +', ' ', output).split(" ")[1]

      if dep_version == "latest":
        # Assume older version so dap will try to upgrade automatically
        return PipInstaller.OLDER_VERSION_INSTALLED
      elif LooseVersion(installed_version) == LooseVersion(dep_version):
        return PipInstaller.SAME_VERSION_INSTALLED
      elif LooseVersion(installed_version) < LooseVersion(dep_version):
        return PipInstaller.OLDER_VERSION_INSTALLED
      elif LooseVersion(installed_version) > LooseVersion(dep_version):
        return PipInstaller.NEWER_VERSION_INSTALLED


class SysInstaller:
  def __init__(self):
    return

  def install(self, dep_name, dep_version="latest", details={}):
    if details and details['env_vars']:
      for env_var_name, env_var_value in details['env_vars'].items():
        os.environ[env_var_name] = str(env_var_value)
        common.print_verbose("set " + str(env_var_name) + "=" + str(env_var_value))
    install_command = [self.installer(), "install",
                       self.dependency_with_version(dep_name, dep_version, self.installer())]
    common.print_verbose(install_command)
    return common.run_command(install_command)

  def dependency_with_version(self, dep_name, dep_version, sys_installer):
    if dep_version == "latest":
      return dep_name
    elif sys_installer == "brew":
      return dep_name + "@" + dep_version
    elif sys_installer == "apt-get":
      return dep_name + "=" + dep_version
    elif sys_installer == "yum":
      return dep_name + "-" + dep_version
    else:
      return ""

  def installer(self):
    if sys.platform.startswith('darwin'):
      return 'brew'
    elif sys.platform.startswith('linux'):
      if self.linux_distribution().startswith('debian'):
        return 'apt-get'
      elif self.linux_distribution().startswith('rhel'):
        return 'yum'
      else:
        common.print_error(
          "Cannot install system dependencies because /etc/os-release does not exist. It is required to determine linux distribution.")
    else:
      common.exit_with_error_message("Sorry, only Linux (Ubuntu, CentOS) and Mac OS X are currently supported")

  def linux_distribution(self):
    exit_code, output = common.run_command([which('grep'), 'ID_LIKE', '/etc/os-release'])
    return output.split("=")[1].rstrip()




#!/usr/bin/env python

#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

import os, sys, subprocess, inspect, yaml, shutil


LOG_LEVEL = "info"
FAILED = 1
SUCCESS = 0
saved_exit_code = SUCCESS
saved_meta_model = None


def run_command(command):
    """Run the specified command"""

    subprocess_exit_code = SUCCESS
    output = ""
    if not str(command):
        print_error("Trying to run None")

    print_verbose("Running command " + str(command))
    try:
        output = subprocess.check_output(command).decode("utf-8")
        if output:
            print_verbose("Output:\n" + output)
    except FileNotFoundError as err:
        subprocess_exit_code = err.errno
        output = err.strerror
        print_error("Exit code: " + str(subprocess_exit_code))
        print_error(output)
    except subprocess.CalledProcessError as err:
        subprocess_exit_code = err.returncode
        print_error("Exit code: " + str(subprocess_exit_code))
        if output:
            print_error(output)
    return subprocess_exit_code, output


def run_command_in_shell_without_output(command):
  """Run the specified command but do not show any output or errors unless you are in verbose mode"""

  subprocess_exit_code = SUCCESS
  output = ""
  if not str(command):
    print_error("Trying to run None")

  print_verbose("Running command " + str(command))
  try:
    output = subprocess.check_output(command, shell=True).decode("utf-8")
    if output:
      print_verbose("Output:\n" + output)
  except FileNotFoundError as err:
    subprocess_exit_code = err.errno
    output = err.strerror
    print_verbose("Exit code: " + str(subprocess_exit_code))
    print_verbose(output)
  except subprocess.CalledProcessError as err:
    subprocess_exit_code = err.returncode
    print_verbose("Exit code: " + str(subprocess_exit_code))
    if output:
      print_verbose(output)
  return subprocess_exit_code, output


def action_name(a):
  return a.name() if callable(a.name) else a.name


def run_command_in_shell(command):
  """Run the specified command in a shell"""

  subprocess_exit_code = SUCCESS
  output = ""
  if not str(command):
    print_error("Trying to run None")

  print_verbose("Running command " + str(command))
  try:
    output = subprocess.check_output(command, shell=True).decode("utf-8")
    if output:
      print_verbose(output)
  except FileNotFoundError as err:
    subprocess_exit_code = err.errno
    output = err.strerror
    print_error("Exit code: " + str(subprocess_exit_code))
    print_error(output)
  except subprocess.CalledProcessError as err:
    subprocess_exit_code = err.returncode
    output = err.stdout
    print_error("Exit code: " + str(subprocess_exit_code))
    if output:
      print_error(output.decode())

  try:
    output = output.decode()
  except (UnicodeDecodeError, AttributeError):
    pass

  return subprocess_exit_code, output


def run_with_io(command):
  subprocess_exit_code = SUCCESS
  output = ""

  if not str(command):
    print_error("Trying to run None")

  print_verbose("Running command " + str(command))
  try:
    process = subprocess.Popen(command)
    output, error = process.communicate()
    if output:
      print_verbose(output)
  except FileNotFoundError as err:
    subprocess_exit_code = err.errno
    output = err.strerror
    print_error("Exit code: " + str(subprocess_exit_code))
    print_error(output)
  except subprocess.CalledProcessError as err:
    subprocess_exit_code = err.returncode
    output = err.stdout
    print_error("Exit code: " + str(subprocess_exit_code))
    if output:
      print_error(output.decode())

  try:
    output = output.decode()
  except (UnicodeDecodeError, AttributeError):
    pass

  return subprocess_exit_code, output


def print_warning(warning):
    global LOG_LEVEL
    if LOG_LEVEL == "verbose" or LOG_LEVEL == "info" or LOG_LEVEL == "warning":
        print_raw("WARNING: " + str(warning))


def print_error(error):
    if LOG_LEVEL == "verbose" or LOG_LEVEL == "info" or LOG_LEVEL == "warning" or LOG_LEVEL == "error":
        print_raw("ERROR: " + str(error))


def print_info(info):
    if LOG_LEVEL == "verbose" or LOG_LEVEL == "info":
        print_raw("INFO: " + str(info))


def print_raw(s):
    print(s, flush=True)


def print_info_no_eol(info):
    sys.stdout.write(info)


def print_verbose(verbose_info):
    if LOG_LEVEL == "verbose":
        print_raw("VERBOSE: " + str(verbose_info))


def print_verbose_no_eol(verbose_info):
    if LOG_LEVEL == "verbose":
        sys.stdout.write(verbose_info)


def stop_if_failed(subprocess_exit_code=SUCCESS, error_message=""):
    if subprocess_exit_code != SUCCESS:
        print_error("FAILED " + error_message)
        sys.exit(FAILED)


def continue_if_failed(subprocess_exit_code=SUCCESS, error_message=""):
    global saved_exit_code
    if subprocess_exit_code != SUCCESS:
        saved_exit_code = FAILED


def exit_code():
    return saved_exit_code


def set_log_level(log_level_to_set):
    global LOG_LEVEL
    LOG_LEVEL = log_level_to_set


def set_meta_model(mm):
    global saved_meta_model
    saved_meta_model = mm
    print_verbose("Meta Model: " + str(saved_meta_model))


def meta_model():
    global saved_meta_model
    if saved_meta_model is None:
      saved_meta_model = configured_meta_model()
    return saved_meta_model


def configured_meta_model():
    configuration = config()
    if 'power_daps_meta_model' in configuration:
      return str(configuration['power_daps_meta_model'])
    return os.getenv("POWER_DAPS_META_MODEL", "power_daps/python3")


def config():
    config_file_location = "config/dap_config.yml"
    configuration = {}
    if os.path.exists(config_file_location):
      config_file_contents = ""
      try:
        with open(config_file_location) as f:
          config_file_contents = f.read()
        f.closed
      except (OSError, IOError) as e:
        print_warning("Config file could not be read from " + config_file_location)
      configuration = yaml.load(config_file_contents, Loader=yaml.SafeLoader)
    return configuration


def power_daps_dir():
    return os.path.dirname(os.path.abspath(__file__)) + "/../../../../"


def app_dir():
    return power_daps_dir() + "apps/dap/"


def dependencies_file_location():
    return os.getcwd() + "/dependencies.yml"


def actions_file_location():
    default_actions_file_location = power_daps_dir() + "actions.yml"
    local_actions_file_location = os.getcwd() + "/actions.yml"
    if os.path.exists(local_actions_file_location):
      return local_actions_file_location
    else:
      return default_actions_file_location


def actions_dir():
    ret_val = os.path.realpath(os.path.abspath(os.path.join(os.path.split(inspect.getfile( inspect.currentframe() ))[0],"../../meta_models/" + meta_model() + "/src")))
    return ret_val


def exit_with_error_message(error_message):
    print_error(error_message)
    sys.exit(1)


def dirs_in(dirname, ignore_dirs=[]):
  subfolders = [f.path for f in os.scandir(dirname) if f.is_dir()]
  subfolders_without_ignore_dirs = []
  for dirname in list(subfolders):
    is_ignore_dir = False
    for ignore_dir in ignore_dirs:
      if ignore_dir in dirname:
        is_ignore_dir = True

    if not is_ignore_dir:
      subfolders_without_ignore_dirs.append(dirname)

    subfolders_without_ignore_dirs.extend(dirs_in(dirname, ignore_dirs))

  return subfolders_without_ignore_dirs


def is_installed(command_name):
  if shutil.which(command_name):
    return True
  return False


def stop_if_not_installed(command_name, additional_error_message=""):
  if is_installed(command_name):
    return
  else:
    error_message = "'" + command_name + "' not found in PATH. " + additional_error_message
    print_error("FAILED " + error_message)
    sys.exit(FAILED)


def which(command_name):
  stop_if_not_installed(command_name)
  return shutil.which(command_name)


def find_files_containing(search_string):
  grep_files_command = [shutil.which('find'), ".", "!", "-name", '*.pyc', "!", "-path", '*.git*', "-type", "f", "-exec",
                        shutil.which("grep"), "-l", search_string, '{}', ";", "-print"]
  return list(set(run_command(grep_files_command)[1].splitlines()))


if __name__ == '__main__':
    print_error("This module " + __file__ + " cannot be run as a stand alone command")
    sys.exit(1)




#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

import os, sys, pathlib, shutil
from dap_core import common
from dap_core.meta_model import MetaModel


def copy_template_files_to(destination, source_action_name):
  common.print_verbose("Looking for files to copy in: " + str(pathlib.Path(MetaModel(common.meta_model()).template_for_action(source_action_name))))
  files_to_copy = [str(p) for p in pathlib.Path(MetaModel(common.meta_model()).template_for_action(source_action_name)).glob("*")]
  common.print_verbose("Found " + str(len(files_to_copy)) + " files to copy.")
  command_to_run = ['/bin/cp', "-R", *files_to_copy, destination]
  common.run_command(command_to_run)


def find_and_replace_in_file_names_and_content(dir, find_and_replace_dict):
  current_dir = os.getcwd()

  for str_to_find, str_to_replace_with in find_and_replace_dict.items():
    dirs = sorted(common.dirs_in(dir, ["__pycache__", "dist", "build", "egg-info", ".git"]) + ["."], key=len, reverse=True)

    for d in dirs:
      os.chdir(current_dir + "/" + d)

      files_to_rename = [str(p) for p in pathlib.Path(".").glob("*" + str_to_find + "*")]

      for f in files_to_rename:
        common.print_verbose("Renaming " + f + " to " + f.replace(str_to_find, str_to_replace_with))
        rename_command = ['/bin/mv', f, f.replace(str_to_find, str_to_replace_with)]
        common.run_command(rename_command)

    grep_files_command = [shutil.which('find'), ".", "!", "-name", '*.pyc', "!", "-path", '*.git*', "-type", "f", "-exec", shutil.which("grep"), "-l", "PROJECT_NAME", '{}', ";", "-print"]
    files_to_search_and_replace_within = common.run_command(grep_files_command)[1].splitlines()

    for f in files_to_search_and_replace_within:
      sed_command = sed_find_and_replace_command(str_to_find, str_to_replace_with, f)
      common.run_command(sed_command)

  os.chdir(current_dir)


def sed_find_and_replace_command(str_to_find, str_to_replace_with, filename):
  sed_command = [shutil.which('sed'), '-i']
  if sys.platform.startswith('darwin'):
    sed_command += [""]
  sed_command += ['-e', "s/" + str_to_find + "/" + str_to_replace_with + "/g", filename]
  return sed_command


def check_that_name_does_not_have_dashes(name):
  if "-" in name:
    common.exit_with_error_message("Name " + name + " has dashes. Please use underscores as dashes cause problems in the python ecosystem")





#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.
import os, re, io
from xml.etree import ElementTree

from dap_core import common
import urllib.request

BASE_URL = "https://repo1.maven.org/maven2/"

class MavenCentralInstallerOld:
  # https://search.maven.org/remotecontent?filepath=
  def __init__(self, base_url=BASE_URL, lib_dir="lib"):
    self.base_url = base_url
    self.lib_dir = lib_dir
    self.latest_versions_cache = {}
    return

  def install(self, name, version, details):
    if not self.has_already_been_downloaded(details["group_id"], name, version, "jar"):
      self.install_file(name, version, details, "jar")
      self.install_file(name, version, details, "pom")
      self.install_transitive_dependencies(name, version, details, "pom")

  def install_transitive_dependencies(self, name, version, details, extension):
    if self.local_pom_exists(details["group_id"], name, version):
      namespaces = {'xmlns': 'http://maven.apache.org/POM/4.0.0'}
      tree = ElementTree.parse(self.local_location(details["group_id"], name, version, "pom"))
      root = tree.getroot()

      deps = root.findall("./xmlns:dependencies/xmlns:dependency", namespaces=namespaces)
      for d in deps:
        version = "latest"
        groupId = d.find("xmlns:groupId", namespaces=namespaces).text
        artifactId = d.find("xmlns:artifactId", namespaces=namespaces).text
        version_elem = d.find("xmlns:version", namespaces=namespaces)
        if version_elem is not None:
          if self.value_references_variable(version_elem.text):
            version_var_name = re.findall("\$\{(.*?)\}", version_elem.text)[0]
            common.print_verbose("Looking for property " + version_var_name)
            props = root.findall("./xmlns:properties", namespaces=namespaces)
            for el in props[0].iter():
              if el.tag == "{http://maven.apache.org/POM/4.0.0}" + version_var_name:
                version = el.text
                common.print_verbose("Found property " + version_var_name + " = " + version)
          else:
            version = version_elem.text
        self.install(artifactId, version, {"group_id": groupId})

  def value_references_variable(self, value):
    return value.startswith("${")

  def install_file(self, name, version, details, extension):
    remote_loc = self.remote_location(details["group_id"], name, version, extension)
    local_lib_dir = self.local_lib_directory(details["group_id"], name, version)
    local_loc = self.local_location(details["group_id"], name, version, extension)
    if not self.has_already_been_downloaded(details["group_id"], name, version, extension):
      common.print_info("Downloading " + remote_loc + " to " + local_loc)
      common.run_command(["mkdir", "-p", local_lib_dir])
      self.fetch(remote_loc, local_loc)
    else:
      common.print_verbose("Dependency found at " + local_loc)
    return 0, ""

  def remote_location(self, group_id, artifact_id, version, file_extension):
    group_id_with_slashes = group_id.replace(".", "/")

    if version != "latest":
      return self.base_url + "/".join(
        [group_id_with_slashes, artifact_id, version, artifact_id]) + "-" + version + "." + file_extension
    else:
      metadata_file = self.metadata_local_location(group_id, artifact_id)
      # TODO: If metadata file is present, don't fetch again
      metadata_url = self.base_url + "/".join([group_id_with_slashes, artifact_id]) + "/maven-metadata.xml"
      self.fetch(metadata_url, metadata_file)
      # TODO: Parse metadata file and return the latest version.
      namespaces = {'xmlns': ''}
      tree = ElementTree.parse(metadata_file)
      root = tree.getroot()
      latest_version = root.find(".//latest").text
      self.add_latest_version_to_cache(group_id, artifact_id, version, file_extension, latest_version)
      return self.base_url + "/".join(
        [group_id_with_slashes, artifact_id, latest_version, artifact_id]) + "-" + latest_version + "." + file_extension

  def add_latest_version_to_cache(self, group_id, artifact_id, version, file_extension, latest_version):
    self.latest_versions_cache["_".join([group_id, artifact_id, version, file_extension])] = latest_version

  def get_latest_version_from_cache(self, group_id, artifact_id, version, file_extension):
    key = "_".join([group_id, artifact_id, version, file_extension])
    if key in self.latest_versions_cache:
      return self.latest_versions_cache[key]
    else:
      return "latest"

  def metadata_local_location(self, group_id, artifact_id):
    group_id_with_slashes = group_id.replace(".", "/")
    return "lib/java/" + "/".join([group_id_with_slashes, artifact_id]) + \
           "/" + "maven-metadata.xml"

  def local_location(self, group_id, artifact_id, version, file_extension):
    v = version
    if version == "latest":
      v = self.get_latest_version_from_cache(group_id, artifact_id, version, file_extension)

    return self.local_lib_directory(group_id, artifact_id, v) + \
           "/" + artifact_id + "-" + v + "." + file_extension

  def local_lib_directory(self, group_id, artifact_id, version):
    group_id_with_slashes = group_id.replace(".", "/")
    return "lib/java/" + "/".join([group_id_with_slashes, artifact_id, version])

  def fetch(self, remote_loc, local_loc):
    urllib.request.urlretrieve(remote_loc, local_loc)

  def has_already_been_downloaded(self, group_id, artifact_id, version, file_extension):
    return os.path.exists(self.local_location(group_id, artifact_id, version, file_extension))

  def local_pom_exists(self, group_id, artifact_id, version):
    return os.path.exists(self.local_location(group_id, artifact_id, version, "pom"))


class MavenCentralInstaller:
  def __init__(self, base_url=BASE_URL, lib_dir="lib"):
    self.base_url = base_url
    self.lib_dir = lib_dir
    self.latest_versions_cache = {}
    return

  def install(self, artifact_id, version, details):
    group_id = details["group_id"]
    jar_dependency = JarDependency(group_id, artifact_id, version)

    jar_dependency.install(self.base_url)


class MavenCentralArtifact:
  def __init__(self, group_id, artifact_id, version):
    self.group_id = group_id
    self.artifact_id = artifact_id
    self.version = version
    self.latest_versions_cache = {}

  def latest_version_from_metadata(self):
    latest_version_from_cache = self.get_latest_version_from_cache()
    if latest_version_from_cache != "latest":
      return latest_version_from_cache
    else:
      tree = ElementTree.parse(self.metadata_file())
      root = tree.getroot()
      latest_version = root.find(".//latest").text
      self.add_latest_version_to_cache(latest_version)

      return latest_version

  def metadata_file(self):
    metadata_file = self.metadata_local_location(self.group_id, self.artifact_id)
    if os.path.exists(metadata_file):
      return metadata_file
    else:
      group_id_with_slashes = self.group_id.replace(".", "/")
      metadata_url = self.base_url + "/".join([group_id_with_slashes, self.artifact_id]) + "/maven-metadata.xml"
      self.fetch(metadata_url, metadata_file)
      return metadata_file

  def relative_remote_location(self):
    if not self.file_extension:
      common.print_error("File extension is not set for " + self.artifact_id)
      self.file_extension.length

    group_id_with_slashes = self.group_id.replace(".", "/")

    return "/".join(
      [group_id_with_slashes, self.artifact_id, self.version, self.artifact_id]
    ) + "-" + self.version + "." + self.file_extension

  def local_location(self):
    return self.local_lib_directory() + \
           "/" + self.artifact_id + "-" + self.version + "." + self.file_extension

  def local_lib_directory(self):
    group_id_with_slashes = self.group_id.replace(".", "/")
    return "lib/java/" + "/".join([group_id_with_slashes, self.artifact_id, self.version])

  def add_latest_version_to_cache(self, latest_version):
    self.latest_versions_cache["_".join([self.group_id, self.artifact_id, self.specified_version, self.file_extension])] = latest_version

  def get_latest_version_from_cache(self):
    key = "_".join([self.group_id, self.artifact_id, self.specified_version, self.file_extension])
    if key in self.latest_versions_cache:
      return self.latest_versions_cache[key]
    else:
      return "latest"

  def has_been_installed(self):
    installed = os.path.exists(self.local_location())
    return installed

  def install(self, base_url):
    if not self.has_been_installed():
      common.run_command(["mkdir", "-p", self.local_lib_directory()])
      fetch(base_url + self.relative_remote_location(), self.local_location(), self.error_callback)


class JarDependency:
  def __init__(self, group_id, artifact_id, version):
    self.group_id = group_id
    self.artifact_id = artifact_id
    self.specified_version = version
    self.metadata = JarDependencyMetadata(group_id, artifact_id)
    self.version = self.determine_version()
    self.pom = Pom(group_id, artifact_id, self.version)
    self.jar = Jar(group_id, artifact_id, self.version)

  def determine_version(self):
    if self.specified_version != "latest":
      return self.specified_version
    return self.metadata.latest_version()

  def install(self, base_url):
    if not self.has_been_installed():
      common.print_info("Downloading " + self.group_id + "/" + self.artifact_id + " v" + self.version)
      self.pom.install(base_url)
      self.jar.install(base_url)

      for d in self.dependencies():
        d.install(base_url)

  def has_been_installed(self):
    common.print_verbose("Checking if " + self.group_id + "/" + self.artifact_id + " is installed.")
    common.print_verbose("           v" + self.version)
    installed = self.pom.has_been_installed() and self.jar.has_been_installed()
    if installed:
      common.print_verbose(self.group_id + "/" + self.artifact_id + " v" + self.version + " already installed.")
    else:
      common.print_verbose(self.group_id + "/" + self.artifact_id + " v" + self.version + " not installed.")
    return installed

  def dependencies(self):
    jar_deps = []
    namespaces = {'xmlns': 'http://maven.apache.org/POM/4.0.0'}
    if self.pom.pom_file() is  None:
      return jar_deps

    root = self.pom.pom_tree().getroot()

    deps = root.findall("./xmlns:dependencies/xmlns:dependency", namespaces=namespaces)
    for d in deps:
      groupId = d.find("xmlns:groupId", namespaces=namespaces).text
      if groupId.startswith("${"):
        groupId = self.pom.resolve_property(groupId)

      artifactId = d.find("xmlns:artifactId", namespaces=namespaces).text
      version = "latest"

      version_elem = d.find("xmlns:version", namespaces=namespaces)
      if version_elem is not None:
        version = version_elem.text

      if version.startswith("${"):
        version = self.pom.resolve_property(version)

      scope_elem = d.find("xmlns:scope", namespaces=namespaces)
      if scope_elem is not None:
        scope = scope_elem.text
      else:
        scope = "unspecified"

      common.print_verbose(self.group_id + "/" + self.artifact_id + " v" + self.version + " depends on " + groupId + "/" + artifactId + " version " + version + " in scope " + scope)
      if scope in ["unspecified", "compile", "runtime"]:
        dep = JarDependency(groupId, artifactId, version)
        jar_deps.append(dep)
        
    return jar_deps

  def __eq__(self, other):
    """Overrides the default implementation"""
    if isinstance(self, other.__class__):
      return self.group_id == other.group_id and self.artifact_id == other.artifact_id and self.version == other.version

    return NotImplemented

  def __hash__(self):
    """Overrides the default implementation"""
    return hash(tuple(sorted([self.group_id, self.artifact_id, self.version])))

  def __repr__(self):
    return "JarDependency[" + self.group_id + "," + self.artifact_id + "," + self.version + "]"


class JarDependencyMetadata:

  def __init__(self, group_id, artifact_id):
    self.group_id = group_id
    self.artifact_id = artifact_id
    self.latest_versions_cache = {}
    self.base_url = BASE_URL

  def latest_version(self):
    latest_version_from_cache = self.get_latest_version_from_cache()
    if latest_version_from_cache != "latest":
      return latest_version_from_cache
    else:
      tree = ElementTree.parse(self.metadata_file())
      root = tree.getroot()
      latest_version = "cannot find"
      latest_version_element = root.find(".//latest")
      version_element = root.find(".//version")
      if latest_version_element is not None:
        latest_version = latest_version_element.text
      elif version_element is not None:
        latest_version = version_element.text

      self.add_latest_version_to_cache(latest_version)
      common.print_verbose("Determined version of " + self.group_id + "/" + self.artifact_id + " to be " + latest_version + " from metadata")
      return latest_version

  def metadata_file(self):
    metadata_file = self.metadata_local_location()
    if os.path.exists(metadata_file):
      return metadata_file
    else:
      common.run_command(["mkdir", "-p", self.metadata_local_directory()])
      fetch(self.metadata_remote_location(), self.metadata_local_location())
      return metadata_file

  def metadata_local_directory(self):
    group_id_with_slashes = self.group_id.replace(".", "/")
    return "lib/java/" + "/".join([group_id_with_slashes, self.artifact_id]) + "/"

  def metadata_local_location(self):
    return self.metadata_local_directory() + "maven-metadata.xml"

  def metadata_remote_location(self):
    group_id_with_slashes = self.group_id.replace(".", "/")
    return self.base_url + "/".join([group_id_with_slashes, self.artifact_id]) + "/maven-metadata.xml"

  def add_latest_version_to_cache(self, latest_version):
    self.latest_versions_cache["_".join([self.group_id, self.artifact_id, "latest"])] = latest_version

  def get_latest_version_from_cache(self):
    key = "_".join([self.group_id, self.artifact_id, "latest"])
    if key in self.latest_versions_cache:
      return self.latest_versions_cache[key]
    else:
      return "latest"


class Pom(MavenCentralArtifact):

  def __init__(self, group_id, artifact_id, version):
    super().__init__(group_id, artifact_id, version)
    self.file_extension = "pom"
    self.namespaces = {'xmlns': 'http://maven.apache.org/POM/4.0.0'}
    self.tree = ""
    self.error_callback = common.print_warning

  def pom_file(self):
    if os.path.exists(self.local_location()):
      return self.local_location()
    else:
      common.print_warning("POM for " + self.group_id + "/" + self.artifact_id + " v" + self.version + " was not found locally. Assuming it has no dependencies.")
      return None

  def resolve_property(self, property_name):
    prop_name = property_name.strip("${}")
    prop_value = "dunno"
    root = self.pom_tree().getroot()

    common.print_verbose("Looking for property " + prop_name + " in " + self.local_location())

    if prop_name.startswith("${project.") or prop_name.startswith("project."):
      common.print_verbose("Looking for project property " + prop_name)
      project_property = prop_name.split(".")[1].rstrip("}")
      prop_value = self.project_property(project_property)
      common.print_verbose("Project property " + prop_name + " = " + prop_value)
      return prop_value
    else:
      props = root.findall("./xmlns:properties", namespaces=self.namespaces)
      if len(props) > 0:
        for el in props[0].iter():
          if el.tag == "{http://maven.apache.org/POM/4.0.0}" + prop_name:
            prop_value = el.text
            common.print_verbose("Found property " + prop_name + " = " + prop_value)
            if prop_value.startswith("${"):
              return self.resolve_property(prop_value)
            else:
              return prop_value
    common.print_verbose("Property " + prop_name + " is not defined in pom. Looking at parent.")
    parent_pom = self.parent_pom()
    if parent_pom:
      prop_value = parent_pom.resolve_property(prop_name)
      common.print_verbose("Parent says " + prop_name + " is " + prop_value)
    else:
      common.print_verbose("No parent found?!")
    return prop_value

  def pom_tree(self):
    if not self.tree:
      self.tree = ElementTree.parse(self.pom_file())
    return self.tree

  def project_property(self, prop_name):
    root = self.pom_tree().getroot()
    # find the project version - need to figure out the xpath
    props = root.findall(".", namespaces=self.namespaces)
    if len(props) > 0:
      for el in props[0].iter():
        if el.tag == "{http://maven.apache.org/POM/4.0.0}" + prop_name:
          prop_value = el.text
          common.print_verbose("Found property " + prop_name + " = " + prop_value)
          return prop_value

      return "no clue"
    return "don't know"

  def parent_pom(self):
    root = self.pom_tree().getroot()
    parents = root.findall("./xmlns:parent", namespaces=self.namespaces)

    if not parents:
      common.print_verbose("No parent found in " + self.local_location())
      return None

    parent_group_id = ""
    parent_artifact_id = ""
    parent_version = ""

    for el in parents[0].iter():
      if el.tag == "{http://maven.apache.org/POM/4.0.0}" + "groupId":
        parent_group_id = el.text
      elif el.tag == "{http://maven.apache.org/POM/4.0.0}" + "artifactId":
        parent_artifact_id = el.text
      elif el.tag == "{http://maven.apache.org/POM/4.0.0}" + "version":
        parent_version = el.text

    if parent_group_id:
      parent = Pom(parent_group_id, parent_artifact_id, parent_version)
      parent.install("https://repo1.maven.org/maven2/")
      return parent
    else:
      return None


class Jar(MavenCentralArtifact):

  def __init__(self, group_id, artifact_id, version):
    super().__init__(group_id, artifact_id, version)
    self.file_extension = "jar"
    self.error_callback = common.exit_with_error_message


def fetch(remote_location, local_location, error_callback=common.exit_with_error_message):
  try:
    common.print_verbose("Fetching " + remote_location + " to " + local_location)
    urllib.request.urlretrieve(remote_location, local_location)
  except urllib.error.HTTPError as err:
    error_callback(str(err.code) + " - Could not retrieve " + remote_location)


#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

import os, inspect, sys, importlib, glob, pkg_resources, pathlib, sysconfig, site
from dap_core import common


class MetaModel:
  n = ""

  def __init__(self, name="power_daps/python3"):
    if '/' not in name:
      name = "power_daps/" + name
    self.n = name

  def name(self):
    return self.n

  def package_name(self):
    pack_name = self.name().replace("/", ".")
    return pack_name

  def load_actions_from_dir(self, dir):

    if os.path.isdir(dir) is not True:
      common.exit_with_error_message("Meta-model '" + self.name() + "' not found in '" + dir + "'")

    elif os.path.isdir(os.path.join(dir, "actions")) is not True:
      common.exit_with_error_message("Meta-model '" + self.name() + "' found but no actions found")

    elif len(self.actions_found_in(dir + "/actions")) == 0:
      common.exit_with_error_message("No actions found in '" + dir + "/actions'")

    if dir not in sys.path:
      sys.path.insert(0, dir)

    actions = []
    #for action in ["default", "deps", "unit_test", "package", "run"]:

    for action in self.actions_found_in(dir + "/actions"):
      sys.path.append(str(pathlib.Path(dir).parent.parent.absolute()))
      action_module = importlib.import_module(self.package_name() + ".actions." + action + "_action")
      actions.append(action_module.action())
    
    return actions


  def actions_dir(self):
    discovered_plugins = {
      entry_point.name: entry_point.load()
      for entry_point
      in pkg_resources.iter_entry_points('power_daps.meta_model.actions')
    }
    # common.print_error(discovered_plugins)
    ret_val = ""
    if not discovered_plugins:
      ret_val = os.path.realpath(
        os.path.abspath(
          os.path.join(
            os.path.split(
              inspect.getfile(
                inspect.currentframe()
              ))[0],
            "../../../meta_models/" + self.name() + "/src/" + self.name())))
    else:
      in_system_sitepackages = sysconfig.get_paths()["purelib"] + "/" + self.name()
      in_user_sitepackages = site.getusersitepackages() + "/" + self.name()
      if os.path.isdir(in_user_sitepackages):
        ret_val = in_user_sitepackages
      elif os.path.isdir(in_system_sitepackages):
        ret_val = in_system_sitepackages
      else:
        ret_val = "."
    return ret_val

  def actions(self):
    return self.load_actions_from_dir(self.actions_dir())

  def actions_found_in(self, dir):
    return [os.path.split(path)[-1].replace("_action.py", "") for path in glob.glob(dir + "/*_action.py")]

  def template_for_action(self, action_name):
    return os.path.join(self.actions_dir(), "templates", action_name)




#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

import pathlib
from dap_core import common
from dap_core.base_actions.dap_action_base import DapActionBase


class CleanActionBase(DapActionBase):
  default_dirs_to_delete = [
    "**/target",
    "**/dist",
    "**/build",
    "**/__pycache__",
    "**/*.egg-info"
  ]

  def __init__(self, additional_dirs_to_delete=[]):
    super().__init__()
    self.dirs_to_delete = CleanActionBase.default_dirs_to_delete + additional_dirs_to_delete
    return

  def run(self):
    super().run()
    for dir_to_delete in self.dirs_to_delete:
      self.delete_dir_named(dir_to_delete)
    return 0, ""

  def delete_dir_named(self, dir_name):
    common.print_verbose("Found " + str(len(self.dirs_named(dir_name))) + " dirs named " + dir_name)

    for d in self.dirs_named(dir_name):
      common.print_verbose("Deleting dir " + d)
      common.stop_if_failed(*common.run_command(["/bin/rm", "-rf", d]))

  def dirs_named(self, dir_name):
    return [str(p.absolute()) for p in pathlib.Path(".").glob(dir_name)]



#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.
#

import sys, platform, yaml
from shutil import which
from dap_core import common
from dap_core.dependency_installers import CommandLineInstaller, PipInstaller, SysInstaller
from dap_core.jar_dependency_installer import MavenCentralInstaller


class Dependencies():
  def __init__(self, dependencies_file_contents):
    self.dependencies_file_contents = dependencies_file_contents
    self.dependencies = dict()
    dependencies_yaml = yaml.load(self.dependencies_file_contents, Loader=yaml.SafeLoader)
    if not dependencies_yaml:
      common.print_verbose("No dependencies found")
      return

    for stage in dependencies_yaml.items():
      stage_name_from_yaml = stage[0]

      # No stages defined
      if stage_name_from_yaml not in self.dependencies:
        self.dependencies[stage_name_from_yaml] = list()

      # Stage is defined but not dependencies listed for the stage
      stage_dependencies_from_yaml = stage[1] if stage[1] else dict()

      for dependency in stage_dependencies_from_yaml.items():
        dep_name = dependency[0]
        dep_version = dependency[1]["version"]
        dep_installer = dependency[1]["installer"]
        dep_details = dict(dependency[1])
        del dep_details["version"]
        del dep_details["installer"]
        self.dependencies[stage_name_from_yaml].append(\
          Dependency(name=dependency[0], version=dep_version, installer=dep_installer, details=dep_details))

    return

  def dependencies_for(self, stage_name):
    return self.dependencies.get(stage_name, list())


class Dependency:

  def __init__(self, name, version, installer, details=dict()):
    self.name = name
    self.version = str(version)
    self.installer_type = installer
    if len(details) > 0:
      self.details = details
    else:
      self.details = None
    return

  def __eq__(self, other):
      """Overrides the default implementation"""
      if isinstance(self, other.__class__):
          return self.__dict__ == other.__dict__
      return NotImplemented

  def __ne__(self, other):
      """Overrides the default implementation (unnecessary in Python 3)"""
      x = self.__eq__(other)
      if x is not NotImplemented:
          return not x
      return NotImplemented

  def __hash__(self):
      """Overrides the default implementation"""
      return hash(tuple(sorted(self.__dict__.items())))

  def installer(self):
    installers = dict()
    installers["npm"] = CommandLineInstaller(['/usr/local/bin/npm', 'install', '--save-dev'])
    installers["pip3"] = PipInstaller()
    installers["system"] = SysInstaller()
    installers["brew_cask"] = CommandLineInstaller([which('brew'), 'cask', 'install'])
    installers["jar"] = MavenCentralInstaller()
    installers["cargo"] = CommandLineInstaller([which('cargo'), 'install'])

    return installers[self.installer_type]

  def install(self):
    common.print_verbose("Installing '" + self.name + "', " + self.version + " version via " + self.installer_type)
    self.installer().install(self.name, self.version, self.details)



#  Copyright 2016-2020 Prasanna Pendse <prasanna.pendse@gmail.com>
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.
#
#  This file is part of power-daps.
#
#  power-daps is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  power-daps is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with power-daps.  If not, see <https://www.gnu.org/licenses/>.

import os
from dap_core import common
from dap_core.util import str_util, template_util, git_util
from dap_core.base_actions.dap_action_base import DapActionBase


class InitActionBase(DapActionBase):

  def __init__(self, pre_copy_commands=[], additional_find_and_replace_dict={}):
    super().__init__()
    self.project_dir = '.'
    self.project_name = os.getcwd().split('/')[-1]

    self.find_and_replace_dict = {
      "PROJECT_NAME": self.project_name,
      "PROJECT_CAMELIZED_NAME": str_util.camelize(self.project_name)}

    additional_find_and_replace_dict.update(self.find_and_replace_dict)
    self.pre_copy_commands = pre_copy_commands

  def run(self):
    super().run()

    template_util.check_that_name_does_not_have_dashes(self.project_name)

    for pre_copy_command in self.pre_copy_commands:
      pre_copy_command.run()

    template_util.copy_template_files_to(self.project_dir, common.action_name(self))
    template_util.find_and_replace_in_file_names_and_content(self.project_dir, self.find_and_replace_dict)
    git_util.git_init_or_add(self.project_dir)
    common.print_raw("Initialized new " + common.meta_model() + " application.")

    return 0, ""
